// Copyright IBM Corp. 2015,2016. All Rights Reserved.
// Node module: wlpn-cli-scaling-member
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

'use strict';

var debug = require('./debug')('scaling-member');
var express = require('express');
var jsonParser = require('body-parser').json({strict: false});
var mandatory = require('./util').mandatory;
var os = require('os');
var urlencode = require('urlencode');


module.exports = ScalingMember;

function ScalingMember(options) {
  this._console = options.logger || console;
  this._manager = mandatory(options.manager, 'manager not provided');
  this.hostName = mandatory(options.hostName, 'hostName not provided');
  this.adminPort = mandatory(options.adminPort, 'adminPort not provided');
  this.serverPath = mandatory(options.serverPath, 'serverPath not provided');
  this.serverName = mandatory(options.serverName, 'serverName not provided');
  
  debug('Host Name: %s Admin Port: %s Server Path: %s Server Name: %s',
         this.hostName, this.adminPort, this.serverPath, this.serverName);
  this._encodedUserDir = urlencode(
    mandatory(options.userDir, 'userDir not provided'));

  this.metricMessageInterval = 30000; // in ms
  this.metricMessageErrorDelay = 90000; // in ms
  this._ErrorDelay = 0; // in ms
  this.timeLastMessageSent = -1;
  this.hostMetrics = {cpu: 0, memory: 0};
  this.hostMetricPercentages = {cpu: 0, memory: 0};
  this.serverMetrics = {cpu: 0, memory: 0, heap: 0};
  this.serverMetricPercentages = {cpu: 0, memory: 0, heap: 0};
  this.hostMax = {cpu: 0, memory: 0};
  this.serverMax = {cpu: 0, memory: 0, heap: 0};
  this.initialized = {
    cpu: false,
    memory: false,
    heap: false,
    all: false,
    sent: false,
  };
  this.hostThresholds = {};
  this.serverThresholds = {};
  this._hasThresholds = {initialized: false};
}

ScalingMember.prototype.start = function(options, callback) {
  debug('start');
  this._controller = this._manager.get('controller')();
  if (!this._controller || !this._controller.request) {
    return callback(new Error('no controller component was found'));
  }

  var self = this;
  var cpus = os.cpus();
  this.hostMax.cpu = 100 * cpus.length;
  this.serverMax.cpu = 100 * cpus.length;
  this.initialized.cpu = true;
  this._console.log('Host Name:%s Admin Port:%s Server Path:%s Server Name:%s',
             this.hostName, this.adminPort, this.serverPath, this.serverName);

  this._app = express();
  this._manager.app.use('/scalingController', this._app);
  this._app.use(jsonParser);
  this._app.use(debug.json.middleware);

  this._app.post('/', function(req, res) {
    self.receiveMessage(self, req.body);
    res.send(); // Note that client ignores the status code.
  });

  debug('setting monitor');
  this._manager.get('runner')().on('request', function(msg) {
    switch (msg.cmd) {
      case 'appmetrics:cpu':
        debug('Updated CPU usage to %j', msg.data);

        // Process and system CPU are percentage, from 0.0 to 1.0.
        var serverCPU = msg.data.process;
        var hostCPU = msg.data.system;

        self.hostMetricPercentages.cpu = hostCPU;
        self.serverMetricPercentages.cpu = serverCPU;
        self.hostMetrics.cpu = hostCPU * self.hostMax.cpu;
        self.serverMetrics.cpu = serverCPU * self.serverMax.cpu;
        break;

      case 'appmetrics:memory':
        debug('Updated Memory usage to %j', msg.data);

        // The total amount of RAM in use on the system in bytes.
        var serverMem = msg.data.physical;
        // The amount of RAM used by the Node.js application in bytes.
        var hostMem = msg.data.physical_used;

        if (!self.initialized.memory) {
          self.initialized.memory = true;
          self.hostMax.memory = msg.data.physical_total;
          self.serverMax.memory = msg.data.physical_total;
        }

        self.hostMetricPercentages.memory = hostMem / self.hostMax.memory;
        self.serverMetricPercentages.memory = serverMem / self.serverMax.memory;
        self.hostMetrics.memory = hostMem;
        self.serverMetrics.memory = serverMem;
        break;

      case 'appmetrics:gc':
        debug('Updated Heap usage to %j', msg.data);

        // The amount of memory used on the JavaScript heap in bytes, only
      // applies to the server, not the host.
        var heapUsage = msg.data.used;

        if (!self.initialized.heap) {
          self.initialized.heap = true;
          self.serverMax.heap = msg.data.size;
        }

        self.serverMetricPercentages.heap = heapUsage / self.serverMax.heap;
        self.serverMetrics.heap = heapUsage;
        break;
    }

    var inited = self.initialized;
    var all = inited.heap && inited.memory && inited.cpu;
    if (!inited.sent && all) {
      self._console.log(
        'ScalingStats.runOnRequest: Sending metrics initialized message.');
      self.initialized.all = true;
      var currentTime = Date.now();
      var timeSinceLastMessage = currentTime - self.timeLastMessageSent;
      if (timeSinceLastMessage >=
          (self.metricMessageInterval + self._ErrorDelay)) {
        var send = self._metricsInitializedMessage(self.hostMax, self.serverMax,
                                                 self.initialized);
        self.timeLastMessageSent = currentTime;
        self.sendMessage(send, function(err) {
          if (!err) {
            self.initialized.sent = true;
            self._ErrorDelay = 0;
          }
          else
            self._ErrorDelay = self.metricMessageErrorDelay;
        });
      };
    };
    self.checkThresholds();
  });
  this._establishConnectionWithParent(callback);
};

ScalingMember.prototype.reconnect = function(server, cb) {
  this._establishConnectionWithParent(cb);
};

/* See if we have any threshold values; else threshold is empty and does not
 * exist.  For example, we could have the following with 0, 1, and 2 thresholds,
 * respectively:
 *
 *   'heap':{'thresholds':[]},
 *   'cpu':{'thresholds':[{'upper':true,'value':90}]},
 *   'memory':{'thresholds':[{'upper':true,'value':90},
 *   {'upper':false,'value':10}]}
 */
function thresholdExists(metric) {
  return metric.thresholds && Object.keys(metric.thresholds).length;
}

// If threshold is initialized, check both the host and server values for
// breaches.
ScalingMember.prototype.checkThresholds = function() {
  var breached = false;
  if (!this._hasThresholds.initialized) {
    this._console.log(
      'ScalingStats.checkThresholds: No thresholds set, returning.');
    return breached;
  }
  var cpuBreach = false;
  var memBreach = false;
  var heapBreach = false;

  // Evaluate thresholds.  First, check if we have a threshold defined for the
  // server or host metric.  Additionally, if server has breached, no need to
  // also check the host as a breach is already found.
  if (this.initialized.cpu) {
    if (this.serverThresholds.cpu && thresholdExists(this.serverThresholds.cpu))
      cpuBreach = compareMetricToThreshold(this.serverMetricPercentages.cpu,
                                           this.serverThresholds.cpu, 'CPU');
    if (this.hostThresholds.cpu && thresholdExists(this.hostThresholds.cpu)) {
      if (!cpuBreach)
        cpuBreach = compareMetricToThreshold(this.hostMetricPercentages.cpu,
                                             this.hostThresholds.cpu, 'CPU');
    }
  }

  if (this.initialized.memory) {
    if (this.serverThresholds.memory &&
      thresholdExists(this.serverThresholds.memory))
      memBreach = compareMetricToThreshold(this.serverMetricPercentages.memory,
                                        this.serverThresholds.memory, 'memory');
    if (this.hostThresholds.memory &&
                                  thresholdExists(this.hostThresholds.memory)) {
      if (!memBreach)
        memBreach = compareMetricToThreshold(this.hostMetricPercentages.memory,
                                             this.hostThresholds.memory,
                                             'memory');
    }
  }

  if (this.initialized.heap) {
    // Heap only applies to the server, not the host.
    if (this.serverThresholds.heap &&
       thresholdExists(this.serverThresholds.heap))
      heapBreach = compareMetricToThreshold(this.serverMetricPercentages.heap,
                                            this.serverThresholds.heap, 'heap');
  }

  // Send message on a breach.
  if (cpuBreach || memBreach || heapBreach) {
    breached = true;
    // Check that we haven't sent a message already in the past 30 seconds.
    var currentTime = Date.now();
    var timeSinceLastMessage = currentTime - this.timeLastMessageSent;
    if (timeSinceLastMessage >=
          (this.metricMessageInterval + this._ErrorDelay)) {
      this._console.log('ScalingStats.checkThresholds: ' +
                        'Breach detected, sending message to controller.');
      var self = this;
      var msg = this._metricsMessage(this.hostMetrics, this.serverMetrics,
                                    this.hostMax, this.serverMax);
      this.timeLastMessageSent = currentTime;
      this.sendMessage(msg, function(err) {
        if (!err)
          self._ErrorDelay = 0;
        else
          self._ErrorDelay = self.metricMessageErrorDelay;
      });
    }
  }
  return breached;
};

function compareMetricToThreshold(metricValue, thresholds, metricName) {
  debug('compareMetricToThreshold: value is %s and thresholds are %s',
          metricValue, JSON.stringify(thresholds));
  var breached = false;

  // Get number of thresholds for this metric
  // (could have upper or lower or both)
  var size = Object.keys(thresholds.thresholds).length;
  var i;
  for (i = 0; i < size; i++) {
    if (thresholds.thresholds[i].upper) { // Check upper threshold
      if ((metricValue * 100) > thresholds.thresholds[i].value) {
        debug('ScalingStats.compareMetricToThreshold :' +
              ' Upper threshold %s breach detected.', metricName);
        breached = true;
      }
    } else { // Check lower threshold
      if ((metricValue * 100) < thresholds.thresholds[i].value) {
        debug('ScalingStats.compareMetricToThreshold :' +
              ' Lower threshold breach detected.', metricName);
        breached = true;
      }
    }
  }
  return breached;
}

ScalingMember.prototype.sendMessage = function(message, callback) {
  // Wrap message for the controller with header information before sending
  var messageToController = this.wrapMessage(message);

  debug('ScalingStats.sendMessage: %j', messageToController);

  var data = {
    path: '/ibm/api/scalingController',
    verb: 'POST',
    data: messageToController,
    qs: undefined,
  };
  this._controller.request(data, callback);
};

ScalingMember.prototype.receiveMessage = function(self, message) {
  debug('processing a message: %s', JSON.stringify(message));
  if (message.messageBody.metrics) {
    self.hostThresholds = message.messageBody.metrics;
    self._console.log('Host thresholds set to: %s',
                  JSON.stringify(self.hostThresholds));
  }
  var fullServerName = 'server/' + self.hostName + ',' +
        self._encodedUserDir + ',' + self.serverName;
  self._console.log('fullServerName: %s', fullServerName);
  if (message.messageBody.containers &&
      message.messageBody.containers[fullServerName].metrics) {
    self.serverThresholds =
            message.messageBody.containers[fullServerName].metrics;
    self._console.log('Server thresholds set to: %s',
                  JSON.stringify(self.serverThresholds));
  }
  if (!self._hasThresholds.initialized) {
    self._hasThresholds.initialized = true;
  }
  self.checkThresholds();
};

ScalingMember.prototype._establishConnectionWithParent = function(callback) {
  var self = this;
  var json = this.initialUpMessage();
  var initialMessage = this.wrapMessage(json);
  var data = {path: '/ibm/api/scalingController', verb: 'POST',
                data: initialMessage, qs: undefined};
  this._console.log('Sending message to scalingController: %s',
                      JSON.stringify(data));
  this._controller.request(data, function(err, res, data) {
    self._console.log('Initialization message sent: err=%s, res=%s, data=%s',
                      err, res, data);
    callback(err, res, data);
  });
};

//
// Messaging methods
//

// Message wrapped for Node - controller communication.
ScalingMember.prototype.wrapMessage = function(message) {
  var wrappedMessage = {
    path: '/ibm/api/scalingController',
    verb: 'POST',
    data: {
      senderConnector: {
        hostName: this.hostName,
        hostPort: String(this.adminPort),
        userDir: this._encodedUserDir,
        serverName: this.serverName,
        type: 'RESTConnector',
      },
      messageBody: message,
    },
  };
  return wrappedMessage;
};

//  Message sent on the start up.
ScalingMember.prototype.initialUpMessage = function() {
  var time = new Date().getTime();
  var containers = {};
  containers['server/' + this.hostName + ',' + this._encodedUserDir +
  ',' + this.serverName] = {
    host: false,
    metrics: {
      heap: {},
      cpu: {},
      memory: {},
    },
    metricsInitialized: false,
    virtualMachine: false,
    sendUpTime: time,
    connector: {
      hostName: this.hostName,
      hostPort: String(this.adminPort),
      userDir: this._encodedUserDir,
      serverName: this.serverName,
      type: 'RESTConnector',
    },
  };
  var hostcontainers = {};
  hostcontainers['host/' + this.hostName] = {
    containers: containers,
    ipAddr: null,
    host: true,
    metrics: {
      cpu: {},
      memory: {},
    },
    metricsInitialized: false,
    virtualMachine: false,
    sendUpTime: time,
  };
  var message = {
    containers: hostcontainers,
  };

  return message;
};

// Message sent on metrics initialized.
ScalingMember.prototype._metricsInitializedMessage = _metricsInitializedMessage;

function _metricsInitializedMessage(hostMax, serverMax, initialized) {
  var time = Date.now();
  var path = ['server/' + this.hostName, this._encodedUserDir, this.serverName]
    .join(',');
  var containers = {};
  containers[path] = {
    host: false,
    metrics: {
      heap: {max: serverMax.heap},
      cpu: {max: serverMax.cpu},
      memory: {max: serverMax.memory},
    },
    metricsInitialized: initialized.all,
    virtualMachine: false,
    sendUpTime: time,
  };

  var hostcontainers = {};

  hostcontainers['host/' + this.hostName] = {
    containers: containers,
    metrics: {
      cpu: {max: hostMax.cpu},
      memory: {max: hostMax.memory},
    },
    metricsInitialized: initialized.all,
    sendUpTime: time,
  };
  var message = {
    containers: hostcontainers,
  };
  return message;
};

// Message sent on threshold breach.
ScalingMember.prototype._metricsMessage = _metricsMessage;

function _metricsMessage(hostUsed, serverUsed, hostMax, serverMax) {
  var time = Date.now();
  var path = ['server/' + this.hostName, this._encodedUserDir, this.serverName]
    .join(',');
  var containers = {};
  containers[path] = {
    host: false,
    metrics: {
      heap: {
        max: serverMax.heap,
        used: serverUsed.heap,
      },
      cpu: {
        max: serverMax.cpu,
        used: serverUsed.cpu,
      },
      memory: {
        max: serverMax.memory,
        used: serverUsed.memory,
      },
    },
    metricsInitialized: true,
    virtualMachine: false,
    sendUpTime: time,
  };
  var hostcontainers = {};
  hostcontainers['host/' + this.hostName] = {
    containers: containers,
    metrics: {
      cpu: {
        max: hostMax.cpu,
        used: hostUsed.cpu,
      },
      memory: {
        max: hostMax.memory,
        used: hostUsed.memory,
      },
    },
    metricsInitialized: true,
    sendUpTime: time,
  };
  var message = {
    containers: hostcontainers,
    _messageSentTime: time,
  };

  return message;
};


